﻿using System;
using TcHmiSrv;
using TcHmiSrv.Management;
using ValueType = TcHmiSrv.ValueType;

namespace ADSCommunicationExtension
{
    // The default class of the server extension must inherit from the IExtension interface.
    // The path of the server extension is available via the ExtensionPath property if the default class of the server extension inherits from the ExtensionWithPath base class instead.
    public class ADSCommunicationExtension : IExtension
    {
        private ITcHmiSrvExtHost _host = null;
        private Context _context = null;

        private object _shutdownLock = new object();
        private bool _isShuttingDown = false;

        private TcHmiSrvRequestListener _requestListener = new TcHmiSrvRequestListener();
        private TcHmiSrvShutdownListener _shutdownListener = new TcHmiSrvShutdownListener();

        private Log _logger = null;

        private Random _rand = new Random("ADSCommunicationExtension".GetHashCode());

        // Initializes the server extension.
        public ErrorValue Init(ITcHmiSrvExtHost host, Context context)
        {
            _logger = new Log(host);

            try
            {
                // Set the host
                _host = host;

                // Copy the specified context and set the local context
                _context = new Context(context);

                // Add event handlers
                _requestListener.OnRequest += OnRequest;
                _shutdownListener.OnShutdown += OnShutdown;

                // Register listeners
                _host.register_listener(context, _requestListener);
                _host.register_listener(context, _shutdownListener);

                _logger.Send(context, "MESSAGE_INIT");
                return ErrorValue.HMI_SUCCESS;
            }
            catch (Exception ex)
            {
                _logger.Send(context, "ERROR_INIT", ex.Message, Severity.Severity_Error);
                return ErrorValue.HMI_E_EXTENSION_LOAD;
            }
        }

        // Called when a client requests a symbol from the extension domain.
        private ErrorValue OnRequest(object sender, TcHmiSrvRequestListener.OnRequestEventArgs e)
        {
            ErrorValue ret = ErrorValue.HMI_SUCCESS;
            Context context = e.Context;
            CommandGroup commands = e.Commands;

            try
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_SUCCESS;
                string mapping = "";

                foreach (Command command in commands)
                {
                    mapping = command.Mapping;

                    try
                    {
                        // Use the mapping to check which command is requested
                        switch (mapping)
                        {
                            case "AdsEditValue":
                                ret = AdsEditValue(command);
                                break;

                            case "WriteAdsValue":
                                ret = WriteAdsValue(command);
                                break;

                            case "ReadAdsValue":
                                ret = ReadAdsValue(command);
                                break;

                            default:
                                ret = ErrorValue.HMI_E_EXTENSION;
                                break;
                        }

                        //if (ret != ErrorValue.HMI_SUCCESS)
                        // Do something on error
                    }
                    catch (Exception ex)
                    {
                        command.ExtensionResult = ExtensionErrorValue.HMI_EXT_FAIL;
                        command.ResultString = _logger.Localize(context, "ERROR_CALL_COMMAND", new string[] { mapping, ex.Message });
                    }
                }
            }
            catch
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_FAIL;
            }
            finally
            {
                if (commands.Result != ExtensionErrorValue.HMI_EXT_SUCCESS)
                {
                    // Reset the read value of the commands to prevent the server from sending invalid data
                    foreach (Command command in commands)
                    {
                        command.ReadValue = null;
                    }

                    ret = ErrorValue.HMI_E_EXTENSION;
                }
            }

            return ret;
        }

        // Called when the server is shutting down. After exiting this function the extension dll will be unloaded.
        private ErrorValue OnShutdown(object sender, TcHmiSrvShutdownListener.OnShutdownEventArgs e)
        {
            // If the extension does not shutdown after 10 seconds (blocking threads) OnShutdown will be called again
            lock (_shutdownLock)
            {
                if (_isShuttingDown)
                {
                    return ErrorValue.HMI_SUCCESS;
                }

                _isShuttingDown = true;

                Context context = e.Context;

                try
                {
                    // Unregister listeners
                    _host.unregister_listener(context, _requestListener);
                    _host.unregister_listener(context, _shutdownListener);

                    _logger.Send(context, "MESSAGE_SHUTDOWN");
                    return ErrorValue.HMI_SUCCESS;
                }
                catch (Exception ex)
                {
                    _logger.Send(context, "ERROR_SHUTDOWN", ex.Message, Severity.Severity_Error);
                    return ErrorValue.HMI_E_EXTENSION;
                }
            }
        }

        // Invokes an ADS sync read request by given data of a command object
        private ErrorValue AdsEditValue(Command command)
        {
            // Get the write value of the command
            Value writeValue = command.WriteValue;

            // Check if the write value is set
            if (writeValue.IsSet)
            {
                // Get the type of the write value
                ValueType type = writeValue.Type;

                // Contains the path of the ADS variable
                string path = string.Empty;

                // Contains the value to send
                Value sendValue = null;

                // Check if the write value is a string
                if (type == ValueType.ValueType_String)
                    path = writeValue;
                // Check if the write value is a map or struct
                else if ((type == ValueType.ValueType_Map) || (type == ValueType.ValueType_Struct))
                {
                    // Set the path of the ADS variable
                    path = writeValue["path"];

                    // Set the value to send
                    sendValue = writeValue["value"];
                }
                else
                    return ErrorValue.HMI_E_TYPE_MISMATCH;

                // Check if the path is null or empty
                if (string.IsNullOrEmpty(path))
                    return ErrorValue.HMI_E_INVALID_PARAMETER;

                // Create a new ADS command
                Command adsCommand = new Command(path);

                // Set the options of the ADS command
                adsCommand.Options = command.Options;

                // Set the write value of the ADS command if the value to send is valid
                if (sendValue != null)
                    adsCommand.WriteValue = sendValue;

                // Create a copy of the local context
                Context context = new Context(_context);

                // Set the domain of the new context to 'ADS'
                context.Domain = "ADS";

                // Execute the ADS command
                ErrorValue ret = _host.exec(context, adsCommand);

                // Check if executing the ADS command failed
                if (ret != ErrorValue.HMI_SUCCESS)
                    return ret;

                // Get the result of the ADS command
                ErrorValue result = adsCommand.Result;

                // Check if the ADS command failed
                if (result != ErrorValue.HMI_SUCCESS)
                    return result;

                // Set the read value of the specified command to the read value of the ADS command
                command.ReadValue = adsCommand.ReadValue;

                // Executing the ADS command failed
                return ErrorValue.HMI_SUCCESS;
            }

            // The write value is not set
            return ErrorValue.HMI_E_INVALID_PARAMETER;
        }

        // Invokes an ADS sync read request and write PLc var to 123
        private ErrorValue WriteAdsValue(Command command)
        {   
            // Contains the path of the ADS variable
            string path = "PLC1.MAIN.nCount";

            // Contains the value to send
            Value sendValue = 123;    

            // Create a new ADS command
            Command adsCommand = new Command(path);

            // Set the options of the ADS command
            adsCommand.Options = command.Options;

            // Set the write value of the ADS command if the value to send is valid
            if (sendValue != null)
                adsCommand.WriteValue = sendValue;

            // Create a copy of the local context
            Context context = new Context(_context);

            // Set the domain of the new context to 'ADS'
            context.Domain = "ADS";

            // Execute the ADS command
            ErrorValue ret = _host.exec(context, adsCommand);

            // Check if executing the ADS command failed
            if (ret != ErrorValue.HMI_SUCCESS)
                return ret;

            // Get the result of the ADS command
            ErrorValue result = adsCommand.Result;

            // Check if the ADS command failed
            if (result != ErrorValue.HMI_SUCCESS)
                return result;

            // Set the read value of the specified command to the read value of the ADS command
            command.ReadValue = adsCommand.ReadValue;

            // Executing the ADS command failed
            return ErrorValue.HMI_SUCCESS;            

            // The write value is not set
            return ErrorValue.HMI_E_INVALID_PARAMETER;
        }

        // Invokes an ADS sync read request and read the current value of nCount from PLC
        private ErrorValue ReadAdsValue(Command command)
        {
            // Contains the path of the ADS variable
            string path = "PLC1.MAIN.nCount";

            // Contains the value to send
            Value sendValue = null;

            // Create a new ADS command
            Command adsCommand = new Command(path);

            // Set the options of the ADS command
            adsCommand.Options = command.Options;

            // Set the write value of the ADS command if the value to send is valid
            if (sendValue != null)
                adsCommand.WriteValue = sendValue;

            // Create a copy of the local context
            Context context = new Context(_context);

            // Set the domain of the new context to 'ADS'
            context.Domain = "ADS";

            // Execute the ADS command
            ErrorValue ret = _host.exec(context, adsCommand);

            // Check if executing the ADS command failed
            if (ret != ErrorValue.HMI_SUCCESS)
                return ret;

            // Get the result of the ADS command
            ErrorValue result = adsCommand.Result;

            // Check if the ADS command failed
            if (result != ErrorValue.HMI_SUCCESS)
                return result;

            // Set the read value of the specified command to the read value of the ADS command
            command.ReadValue = adsCommand.ReadValue;

            // Executing the ADS command failed
            return ErrorValue.HMI_SUCCESS;

            // The write value is not set
            return ErrorValue.HMI_E_INVALID_PARAMETER;
        }
    }
}
